##########################################################################
# This file contains two R routines for converting between Julian and
# calendar dates (intended as substitutes for the Splus routines julian()
# and month.day.year(). The routines are as follows:
#
#       jtodmy          Converts from Julian day numbers to Gregorian
#                       calendar dates.
#       dmytoj          Converts from Gregorian calendar dates to Julian
#                       day numbers. 
# Algorithms obtained from http://hermetic.magnet.ch/cal_stud/jdn.htm
# on 28th November 2001 (Julian day 2452242). R version converted from
# FORTRAN on 24th October 2002 (Julian day 2452572).
##########################################################################

##########################################################################
##########################################################################
##########################################################################
#     FUNCTION jtomdmy
#     ^^^^^^^^^^^^^^^^
#
#     This function converts a date in JULIAN format to DD/MM/YY in
#     the Gregorian calendar. Arguments:
#
#     jdate      Julian day number to be converted
#     y0,m0,d0   Year, month and day of `origin'. Defaults to 24th
#                November, 4713BC (which is when the calendar starts!).
#                The origin is actually 1st January 4712BC according
#                to the *Julian* calendar.
#
#     Value: a list containing 3 components (Year, Month and Day).
#
#     The algorithm used here is due to Henry F. Fliegel and Thomas C.
#     Van Flandern. It is only valid for dates after 1st March, 4900BC.
##########################################################################
jtodmy <- function(jdate,d0=NULL,m0=NULL,y0=NULL) {
  if (is.null(y0) & is.null(m0) & is.null(d0)) {
    jd <- jdate
  } else {
    jd <- dmytoj(d0,m0,y0) + jdate
  }
  l <- jd + 68569
  n <- trunc((4*l) / 146097)
  l <- l - trunc(((146097*n) + 3) / 4)
  i <- trunc((4000 * (l+1)) / 1461001)
  l <- l - trunc((1461*i) / 4) + 31
  j <- trunc((80*l) / 2447)
  dd <- l - trunc((2447*j) / 80)
  l <- trunc(j / 11)
  mm <- j + 2 - (12*l)
  yy = (100 * (n-49)) + i + l
  if (any(((10000*yy)+(100*mm)+dd) < -49000301)) {
    stop("Algorithm only works for dates after 1st March 4900BC!")
  }
  list(year=yy,month=mm,day=dd)
}
##########################################################################
##########################################################################
##########################################################################
#     FUNCTION dmytoj
#     ^^^^^^^^^^^^^^^^
#
#     This function converts Gregorian calendar dates to Julian day
#     numbers (Julian origin is 24th November 4713BC). Arguments:
#
#     dd,mm,yy   Day, month and year to be converted
#
#     Value: the Julian conversion.
#
#     The algorithm used here is due to Henry F. Fliegel and Thomas C.
#     Van Flandern. It is only valid for dates after 1st March, 4800BC.
##########################################################################
dmytoj <- function(dd,mm,yy) {
  if (any(yy == 0)) {
    stop("Gregorian year 0 doesn't exist!")
  } else if (any(((10000*yy)+(100*mm)+dd) < -48000301)) {
    stop("Algorithm only works for dates after 1st March 4800BC!")
  }
  jdate <- trunc(1461 * (yy + 4800 + trunc((mm - 14) / 12)) / 4) +
           trunc(367 * (mm - 2 - 12 * trunc((mm - 14 ) / 12)) / 12) -
           trunc(3 * trunc(((yy + 4900 + trunc((mm - 14) / 12)) / 100)) / 4) +
            dd - 32075
  jdate
}
