ll.to.grid <- function(long,lat,N0=-100000,E0=400000,F0=0.9996012717,
                       phi0=pi*49/180,lambda0=-pi*2/180,
                       a=6377563.396,b=6356256.910) {
######################################################################
# Function to convert latitudes and longitudes to grid coordinates.
# By default the grid is the UK National Grid; however, by changing the
# parameters, other grids can be obtained as well. Arguments:
#
# long    Vector of longitudes, in degrees
# lat     Vector of latitudes, in degrees
# N0      Northing of true grid origin
# E0      Easting of true grid origin
# F0      Scale factor on central meridian
# phi0    Latitude of true origin
# lambda0 Longitude of true origin and central meridian
# a       Length of semi-major axis of projection ellipsoid (m)
# b       Length of semi-minor axis  "     "          "      "
#
# Value: a data frame containing 2 columns named `East' and `North',
# with the translated co-ordinates of the points in long and lat.
#
# Formulae are taken from "A guide to coordinate systems in Great Britain",
# published by the Ordnance Survey.
######################################################################
  e2 <- 1 - (b/a)^2; n <- (a-b)/(a+b)
  phi <- pi*lat/180; lambda <- pi*long/180
  sp <- sin(phi)
  tmp1 <- 1 - (e2*sp^2)
  nu <- a*F0/sqrt(tmp1); rho <- a*F0*(1-e2)/(tmp1^1.5)
  ratio <- (1-(e2*sp^2))/(1-e2); eta2 <- ratio-1
  tmp1 <- phi-phi0; tmp2 <- phi+phi0
  M <- b*F0*( ((1+n+(1.25*n^2*(1+n)))*tmp1) -
              (3*n*(1+n+(7*(n^2)/8))*sin(tmp1)*cos(tmp2)) +
              (15*n^2*(1+n)*sin(2*tmp1)*cos(2*tmp2)/8) -
              (35*n^3*sin(3*tmp1)*cos(3*tmp2)/24) )
  z1 <- M + N0; z2 <- nu*sin(phi)*cos(phi)/2
  z3 <- nu*sin(phi)*(cos(phi)^3)*(5-(tan(phi)^2)+(9*eta2))/24
  z3a <- nu*sin(phi)*(cos(phi)^5)*(61-(58*tan(phi)^2)+(tan(phi)^4))/720
  z4 <- nu*cos(phi)
  z5 <- nu*(cos(phi)^3)*(ratio-(tan(phi)^2))/6
  z6 <- nu*(cos(phi)^5)*(5-(18*(tan(phi)^2)) + (tan(phi)^4) + (14*eta2) -
                         (58*eta2*(tan(phi)^2))) / 120
  tmp1 <- lambda-lambda0
  North <- z1 + (z2*tmp1^2) + (z3*tmp1^4) + (z3a*tmp1^6)
  East <- E0 + (z4*tmp1) + (z5*tmp1^3) + (z6*tmp1^5)
  data.frame(North=North,East=East)
}
######################################################################
######################################################################
######################################################################
grid.to.ll <- function(east,north,N0=-100000,E0=400000,F0=0.9996012717,
                       phi0=pi*49/180,lambda0=-pi*2/180,
                       a=6377563.396,b=6356256.910) {
######################################################################
# Function to convert grid coordinates to latitudes and longitudes.
# Arguments as for ll.to.grid(), except that long and lat are replaced
# by east and north respectively.
#
# Value: a data frame containing 2 columns named `Long' and `Lat',
# with the translated co-ordinates of the points in east and north
# (in degrees)
#
######################################################################
  e2 <- 1 - (b/a)^2; n <- (a-b)/(a+b); M <- Inf
  phi1 <- (north-N0)/(a*F0) + phi0
  while (max(abs(north-N0-M)) >= 1e-4) {
    tmp1 <- phi1-phi0; tmp2 <- phi1+phi0
    M <- b*F0*( ((1+n+(1.25*n^2*(1+n)))*tmp1) -
                (3*n*(1+n+(7*(n^2)/8))*sin(tmp1)*cos(tmp2)) +
                (15*n^2*(1+n)*sin(2*tmp1)*cos(2*tmp2)/8) -
                (35*n^3*sin(3*tmp1)*cos(3*tmp2)/24) )
    phi1 <- (north-N0-M)/(a*F0) + phi1
  }
  sp <- sin(phi1); tmp1 <- 1 - (e2*sp^2)
  nu <- a*F0/sqrt(tmp1); rho <- a*F0*(1-e2)/(tmp1^1.5)
  ratio <- (1-(e2*sp^2))/(1-e2); eta2 <- ratio-1
  T <- tan(phi1)
  z7 <- T/(2*rho*nu)
  z8 <- T*(5+(3*T^2)+eta2 -(9*eta2*T^2))/(24*rho*nu^3)
  z9 <- T*(61+(90*T^2)+(45*T^4))/(720*rho*nu^5)
  sec <- 1/cos(phi1)
  z10 <- sec/nu
  z11 <- sec*(ratio+(2*T^2))/(6*nu^3)
  z12 <- sec*(5+(28*T^2)+(24*T^4))/(120*nu^5)
  z12a <- sec*(61+(662*T^2)+(1320*T^4)+(720*T^6))/(5040*nu^7)
  tmp1 <- east-E0
  Lat <- phi1 - (z7*tmp1^2) + (z8*tmp1^4) - (z9*tmp1^6)
  Long <- lambda0 + (z10*tmp1) - (z11*tmp1^3) + (z12*tmp1^5) - (z12a*tmp1^7)
  data.frame(Long=Long*180/pi,Lat=Lat*180/pi)
}
######################################################################
######################################################################
######################################################################
howfar <- function(sites1,sites2,units="km") {
######################################################################
#
# Function to calculate the distance between sets of points on the 
# earth's surface, with co-ordinates given in latitude and longitude.
# Arguments:
#
# sites1	A matrix with 2 columns, indicating the (lat, long)
#		coordinates of each site in the first set (in 
#		degrees!)
# sites2	The same for the second set of sites
# units		Either "km" (the default) or "nm" for distances in 
#		nautical miles (*not* nanometers!)
#
# Value:	A matrix of distances, with rows corresponding to 
#		the sites in sites1 and columns corresponding to 
#		those in sites2. If sites1 and sites2 have row names,
#		these are copied over as appropriate
######################################################################
 n1 <- dim(sites1)[1]; n2 <- dim(sites2)[1]
 z <- matrix(nrow=n1,ncol=n2)
 rownames(z) <- rownames(sites1) 
 colnames(z) <- rownames(sites2) 
 pi180 <- pi/180
 for (i in 1:n1) {
  lat1 <- sites1[i,1]
  long1 <- sites1[i,2]
  lat2 <- sites2[,1]
  long2 <- sites2[,2]
  z[i,] <- (60*180/pi)*acos(sin(sites1[i,1]*pi180)*sin(sites2[,1]*pi180) + 
                            cos(sites1[i,1]*pi180)*cos(sites2[,1]*pi180) * 
                       cos((sites2[,2]-sites1[i,2])*pi180))
 }
 if (units == "km") {
  z <- 1.852*z
 } else if (units != "nm") {
  stop("units must be either 'km' or 'nm'")
 }
 z
}

