###############################################################################
###############################################################################
###############################################################################
######                                                                   ######
######    Creation of precipitation, temperature and predictor datasets  ######
######    northern Iberia, for downscaling using Rglimclim (case study   ######
######    for VALUE training school in Trieste, November 2014).Note      ######
######    that it takes a while (45 minutes or so) to run this script,   ######
######    so don't run it unless you have to! It is provided mainly      ######
######    for documentation purposes, so that delegates can see          ######
######    exactly what was done.                                         ######
######                                                                   ######
###############################################################################
###############################################################################
###############################################################################
require(Rglimclim); require(ncdf)
source("llconvert.r"); source("dateutil.r")
###############################################################################
#
# Read station information and produce a readable output file
#
###############################################################################
get.stationinfo <- function(inflnm,country.name) {
#
# Reads a metadata file and returns a data frame containing site ID,
# name, latitude, longitude and altitude. "country.name" is used
# to identify the rows in the file corresponding to data values
#
  stations <- scan(inflnm,what=character(0),sep="\n")
  stations <- stations[grep(country.name,stations)]
#
# Transform to a matrix in which rows correspond to stations; and strip
# out the unwanted column containing the country name
#
  stations <- t(simplify2array(strsplit(stations,",")))[,-3]
  stations <- as.data.frame(stations)
  names(stations) <- c("ID","Name","Latitude","Longitude","Altitude")
  stations$Altitude <- gsub(" m ","",stations$Altitude)
#
# Altitude in hundreds of m
#
  stations$Altitude <- as.numeric(stations$Altitude) / 100
#
# Convert the character representations of latitude and longitude to 
# decimal values. NB careful with signs!
#
  tmp <- t(simplify2array(strsplit(as.character(stations$Longitude),":")))
  stations$Longitude <- sign(as.numeric(tmp[,1])) * (
    abs(as.numeric(tmp[,1])) + as.numeric(tmp[,2])/60 + as.numeric(tmp[,3])/3600)
  tmp <- t(simplify2array(strsplit(as.character(stations$Latitude),":")))
  stations$Latitude <- sign(as.numeric(tmp[,1])) * (
    abs(as.numeric(tmp[,1])) + as.numeric(tmp[,2])/60 + as.numeric(tmp[,3])/3600)
#
# And append a leading space to 3-character IDs
#
  stations$ID <- as.character(stations$ID)
  stations$ID[nchar(stations$ID)==3] <- 
    paste(" ",stations$ID[nchar(stations$ID)==3],sep="")
  stations
}
precip.stations <- get.stationinfo("ECA/PRECIP/metadata.txt","SPAIN")
temp.stations <- get.stationinfo("ECA/TEMP/metadata.txt","SPAIN")
if (!isTRUE(all.equal(precip.stations,temp.stations))) {
  stop("Precipitation and temperature have different sets of stations")
}
#
# Latitude and longitude limits chosen to correspond roughly to the 
# region indicated in the VALUE map (stations_NorthIberia.pdf). The 
# southernmost stations there are 1395 at Lat 42.29 and 3954 at
# 42.5. The easternmost stations are 422 at Long -0.35 and 3950 at -0.336.
# Here, retain stations north of 42 degrees and west of 0 degrees (can
# always discard a few later if necessary). This is *probably* the area
# for which I downloaded GTOPO30 data to get mapped altitudes and slope 
# information. NB I can't quite remember this (and the site - 
# http://webmap.ornl.gov/wcsdown/dataset.jsp?ds_id=10003) - is currently 
# down so I can't check it :-() - but the dimensions of the array suggest
# that it's 2 degrees of latitude and 9.5 degrees of longitude. Check this 
# by extracting elevations for stations under this hypothesis, and plotting 
# against the documented station elevations. 
#
LatLims <- c(42,44)
LongLims <- c(-9.5,0)
precip.stations <- precip.stations[precip.stations$Longitude > LongLims[1] &
                                     precip.stations$Longitude < LongLims[2] &
                                     precip.stations$Latitude > LatLims[1] &
                                     precip.stations$Latitude < LatLims[2],]
nsites<- nrow(precip.stations)
topo.file <- open.ncdf("GTOP30_Data.cdf")
topo.data <- get.var.ncdf(topo.file,"Band1")
topo.data[topo.data == -9999] <- NA
close.ncdf(topo.file)
#
# Turn the matrix the right way around and convert to hundreds of metres
#
topo.data <- topo.data[,rev(1:ncol(topo.data))] / 100
#
# GTOPO30 is at 30 arcsecond resolution, which is 1/120 of a degree
#
topo.longs <- seq(-9.5+(1/240),-(1/240),by=1/120)
topo.lats <- seq(42+(1/240),44-(1/240),by=1/120)
nlats <- length(topo.lats); nlongs <- length(topo.longs)
#
# Plot map, and recorded versus mapped altitudes
#
if (dev.cur() == 1) x11(width=8,height=4)
par(mfrow=c(1,1),mar=c(4,4,3,2),mgp=c(2,0.75,0))
image(topo.longs,topo.lats,topo.data,col=terrain.colors(100),
      xlab="Longitude (degrees)",ylab="Latitude (degrees)")
box(lwd=2)
title(main="Northern Iberia study area: longitude -9.5 to 0, latitude 42 to 44",
      cex.main=1)
points(precip.stations$Longitude,precip.stations$Latitude,pch=15)
legend("topleft",pch=15,legend="Stations used in study",bty="n",cex=0.8)
dev.copy(pdf,"TerrainMap.pdf",width=8,height=4)
dev.off()

X0 <- LongLims[1] + 1/240
Y0 <- LatLims[1] + 1/240
cellsize <- 1/120
which.long <- floor((precip.stations$Longitude-X0) / cellsize) + 1
which.lat <- floor((precip.stations$Latitude-Y0) / cellsize) + 1
map.elevs <- topo.data[cbind(which.long,which.lat)]
par(mar=c(3,3,3,2))
plot(precip.stations$Altitude,map.elevs,pch=16,xlab="Recorded",ylab="Mapped",
     main=paste("Site altitudes (hundreds of m) from data documentation",
                "and digital elevation map",sep="\n"),cex.main=1)
box(lwd=2)
dev.copy(pdf,"AltitudeCheck.pdf",width=8,height=4)
dev.off()
#
# OK, that looks fine, so add the mapped altitudes to the station
# info and also compute aspect and slope information for each station.
#
precip.stations$MappedAltitude <- map.elevs
#
# Define slopes by regressing altitudes upon co-ordinates, for regions 
# of roughly 3*3km^2, 10*10km^2 and 30*30km^2 centred on each location
# (corresponding to areas of approximately 10, 100 and 1000km^2). And measures 
# of mean altitude and topographic variability in the neighbourhood of 
# each gauge, on basis of experience with rainfall in the Sekhuhune 
# district of South Africa.
#
cutout.size <- c(3,10,30) # This is in km
EW.aspect <- matrix(nrow=nsites,ncol=length(cutout.size))
NS.aspect <- EW.aspect
Elev.mean <- EW.aspect
Elev.sd <- EW.aspect
EW.pixelsize <- 
  howfar(matrix(c(topo.lats[round(nlats/2)],topo.longs[1]),nrow=1),
         matrix(c(topo.lats[round(nlats/2)],topo.longs[nlongs]),nrow=1)) / 
  (nlongs-1)
NS.pixelsize <- 
  howfar(matrix(c(topo.lats[1],topo.longs[round(nlongs/2)]),nrow=1),
         matrix(c(topo.lats[nlats],topo.longs[round(nlongs/2)]),nrow=1)) / 
  (nlats-1)
dev.off()
x11(width=8,height=5)
opar <- par(no.readonly=TRUE)
par(mfrow=c(length(cutout.size),4),mar=c(2,1,3,1),mgp=c(2,0.75,0))
for (k in 1:length(cutout.size)) {
  cur.cut <- cutout.size[k]
  cat(paste("SLOPE COMPUTATIONS - ",cur.cut,"*",cur.cut,"km^2 ...\n\n",sep=""))
  EW.map <- matrix(nrow=nrow(topo.data),ncol=ncol(topo.data))
  NS.map <- EW.map
  Elev.map <- EW.map
  ElevSD.map <- EW.map 
  Ncutout.EW <- round((cur.cut-EW.pixelsize)/(2*EW.pixelsize))
  Ncutout.NS <- round((cur.cut-NS.pixelsize)/(2*NS.pixelsize))
  E.dist <- rep(1:(2*Ncutout.EW+1),1+2*Ncutout.NS)*EW.pixelsize
  N.dist <- rep(1:(2*Ncutout.NS+1),each=1+2*Ncutout.EW)*NS.pixelsize
  if (length(E.dist) < 2 | length(N.dist) < 2) {
    stop("Cutouts are too small - need at least 2 pixels to compute a slope")
  }
  for (i in (Ncutout.EW+1):(nlongs-Ncutout.EW)) {
    cat(paste("Computing slopes: longitude",i-Ncutout.EW,"of",
              nlongs-(2*Ncutout.EW),"\n"))
    for (j in (Ncutout.NS+1):(nlats-Ncutout.NS)) {
      cur.elev <- topo.data[(i-Ncutout.EW):(i+Ncutout.EW),
                            (j-Ncutout.NS):(j+Ncutout.NS)]
      if (is.na(cur.elev[Ncutout.EW+1,Ncutout.NS+1])) next # Current pixel in sea
      cur.elev <- as.vector(cur.elev)
      cur.elev[is.na(cur.elev)] <- 0 # NAs are ocean
      EW.map[i,j] <- coef(lm(cur.elev ~ E.dist,na.action=na.omit))[2]
      NS.map[i,j] <- coef(lm(cur.elev ~ N.dist,na.action=na.omit))[2]
      ElevSD.map[i,j] <- sd(cur.elev,na.rm=TRUE)
      Elev.map[i,j] <- mean(cur.elev,na.rm=TRUE)
    }
  }
  image(topo.longs,topo.lats,Elev.map,axes=FALSE,xlab="",ylab="",
        col=topo.colors(100),
        main=substitute("Elevation: "*cc%*%cc*km^2,list(cc=cur.cut)))
  points(precip.stations$Longitude,precip.stations$Latitude,pch=20)
  box(lwd=2)
  image(topo.longs,topo.lats,ElevSD.map,axes=FALSE,xlab="",ylab="",
        col=topo.colors(100),
        main=substitute("Topographic variation: "*cc%*%cc*km^2,list(cc=cur.cut)))
  points(precip.stations$Longitude,precip.stations$Latitude,pch=20)
  box(lwd=2)
  image(topo.longs,topo.lats,EW.map,xlab="",ylab="",axes=FALSE,
        col=topo.colors(100),
        main=substitute("E-W slopes: "*cc%*%cc*km^2,list(cc=cur.cut)))
  points(precip.stations$Longitude,precip.stations$Latitude,pch=20)
  box(lwd=2)
  image(topo.longs,topo.lats,NS.map,xlab="",ylab="",axes=FALSE,
        col=topo.colors(100),
        main=substitute("N-S slopes: "*cc%*%cc*km^2,list(cc=cur.cut)))
  points(precip.stations$Longitude,precip.stations$Latitude,pch=20)
  box(lwd=2)
  EW.aspect[,k] <- EW.map[cbind(which.long,which.lat)]
  NS.aspect[,k] <- NS.map[cbind(which.long,which.lat)]
  Elev.mean[,k] <- Elev.map[cbind(which.long,which.lat)]
  Elev.sd[,k] <- ElevSD.map[cbind(which.long,which.lat)]
}
dev.copy(pdf,"NorthernIberia_topovar.pdf",width=8,height=5)
dev.off()
precip.stations <- cbind(precip.stations,Elev.mean,Elev.sd,EW.aspect,NS.aspect)
names(precip.stations)[7:18] <- 
  c("Elev_10km^2","Elev_100km^2","Elev_1000km^2",
    "ElevSD_10km^2","ElevSD_100km^2","ElevSD_1000km^2",
    "E-W_slope_10km^2","E-W_slope_100km^2","E-W_slope_1000km^2",
    "N-S_slope_10km^2","N-S_slope_100km^2","N-S_slope_1000km^2")
#
# Before writing to file, remove any stations for which any attributes
# are missing (typically due to their being too close to the borders 
# of the region for which elevation data were extracted)
#
precip.stations <- precip.stations[rowSums(is.na(precip.stations))==0,]
write.table(precip.stations,file="stations.dat",sep="\t",row.names=FALSE)
cat(paste("Processed station information for",nrow(precip.stations),"sites\n"))
###############################################################################
#
# Now read the data for each response variable, and assemble into a common data 
# frame (NB the station ID from the data file, corresponding to the station 
# information read previously, is embedded in text in line 17: the "STAID"
# and "SOUID" columns do not match with this)
#
###############################################################################
read.data <- function(filename,LatLims,LongLims,station.info) {
  cat(paste("Reading data for file",filename,"...\n"))
  Station.ID <- scan(filename,skip=16,n=1,what=character(0),quiet=TRUE,sep="\n")
  Station.ID <- unlist(strsplit(Station.ID,"STAID:"))[2]
  Station.ID <- gsub(")","",Station.ID)
  if (nchar(Station.ID) > 4) Station.ID <- gsub(" ","",Station.ID)
  which.station <- (station.info$ID == Station.ID)
  if (!any(which.station)) return(NULL)
  cur.lat <- station.info$Latitude[which.station]
  cur.long <- station.info$Longitude[which.station]
  if (cur.lat < LatLims[1] | cur.lat > LatLims[2] |
        cur.long < LongLims[1] | cur.long > LongLims[2]) {
    z <- NULL
  } else {
    cat(paste("Station",Station.ID,"at longitude ",round(cur.long,2),
              "and latitude",round(cur.lat,2),"\n"))
    z <- read.table(filename,sep=",",skip=21,na.strings="-9999",header=TRUE)[,(3:5)]
    z$Scode <- rep(Station.ID,nrow(z))
  }
  z
}

precip.files <- list.files("./ECA/PRECIP/",pattern="^RR")
temp.files <- list.files("./ECA/TEMP/",pattern="^TG")

if(exists("precip.data")) rm(precip.data)
for (file in paste("ECA/PRECIP/",precip.files,sep="")) {
  if (!exists("precip.data")) {
    precip.data <- read.data(file,LatLims,LongLims,precip.stations)
  } else {
    precip.data <- 
      rbind(precip.data,read.data(file,LatLims,LongLims,precip.stations))
  }
}
if(exists("temp.data")) rm(temp.data)
for (file in paste("ECA/TEMP/",temp.files,sep="")) {
  if (!exists("temp.data")) {
    temp.data <- read.data(file,LatLims,LongLims,temp.stations)
  } else {
    temp.data <- 
      rbind(temp.data,read.data(file,LatLims,LongLims,temp.stations))
  }
}
#
# Merge the datasets and do some postprocessing. NB the rainfall values
# are in units of 0.1mm, so convert to mm. Similarly the temperatures
# are in tenths of a degree.
#
# For the postprocessing, mark as missing any value that is recorded
# as suspect (quality flag 1) - basically this means retaining only 
# records with a quality flag of 0.
#
all.data <- merge(precip.data,temp.data)
all.data$RR <- all.data$RR / 10
all.data$RR[all.data$Q_RR != 0] <- NA
all.data$TG <- all.data$TG / 10
all.data$TG[all.data$Q_TG != 0] <- NA
all.data <- all.data[order(all.data$DATE,all.data$Scode),]
all.data$Year <- as.numeric(substr(all.data$DATE,1,4))
all.data$Month <- as.numeric(substr(all.data$DATE,5,6))
all.data$Day <- as.numeric(substr(all.data$DATE,7,8))
all.data <- all.data[,c("Year","Month","Day","Scode","RR","TG")]
write.GLCdata(all.data[all.data$Year %in% 1960:1990,],
                               file="NorthIberiaPrecipTemp_1960-1990.dat")
write.GLCdata(all.data[all.data$Year %in% 1990:2012,],
                               file="NorthIberiaPrecipTemp_1990-2012.dat")
###############################################################################
#
# Finally, the predictor data. Consider a variety of different possibilities
# here. The first are simply based on gridded values of MSLP, surface
# temperature, along with measures of airflow and moisture availability, 
# over the region Z8 considered by Gutierrez et al (2013) - this is from 
# longitude 10W to 15E, latitude 27.5N to 45N. These are daily values from 
# the ERA40 reanalysis (rather than the ERA-INTERIM reanalysis advocated
# by VALUE - the reason being that INTERIM only goes back to 1979 whereas
# the COST733 weather classificiations - see below - stop in 2002. To get
# a fair comparison of classification-based versus predictor-based 
# downscaling, it is helpful to use predictor datasets that are roughly
# contemporary).
#
###############################################################################
#
# Scales / offsets needed to convert to sensible units
#
scales <- c(0.01,1,1,1,1)
offsets<- c(1e5,0,0,273,273)
#
# Date sequences
#
start.jdate <- dmytoj(1,9,1957)
end.jdate <- dmytoj(31,8,2002)
jdates <- start.jdate:end.jdate
all.dates <- jtodmy(jdates)
daily.preds <- data.frame(Year=all.dates$year,Month=all.dates$month,
                          Day=all.dates$day)
Input.file <- open.ncdf("ERA40Predictors.nc")
var.names <- c("msl","u10","v10","t2m","d2m")
for (v in 1:length(var.names)) {
  tmp.data <- get.var.ncdf(Input.file,varid=var.names[v])
  tmp.data[tmp.data == -32767] <- NA
#
# Now convert to *sensible* units
#
  tmp.data <- scales[v] * (tmp.data - offsets[v])
#
# Average over whole area
#
  tmp.data <- colMeans(tmp.data,dims=2)
#
# And now create daily values (probably quickest to copy the vector into
# a 4-row matrix and then get the column means, so as to access low-level
# functions)
#
  daily.preds[,v+3] <- colMeans(matrix(tmp.data,nrow=4))
}    
#
# The ERA40 distribution doesn't give wind speed directly; calculate it here, 
# therefore (NB no need to shift and scale for the wind vector components,
# which makes it easy)
#
tmp.data <- sqrt(get.var.ncdf(Input.file,varid="u10")^2 + 
                   get.var.ncdf(Input.file,varid="v10")^2)
tmp.data <- colMeans(tmp.data,dims=2)
daily.preds[,v+4] <- colMeans(matrix(tmp.data,nrow=4))
#
# Put the windspeed adter the u and v components
#
daily.preds <- daily.preds[,c(1:6,9,7:8)]
names(daily.preds)[-(1:3)] <- c(var.names[1:3],"w10",var.names[4:5])
close.ncdf(Input.file)

monthly.preds <- aggregate(daily.preds,
                           by=list(daily.preds$Month,daily.preds$Year),
                           FUN=mean)[,-c(1:2,5)]

long.names <- c("Mean MSLP for region Z8 (mbar-1000)",
                "Mean u-component of 10m wind for region Z8 (ms^(-1))",  
                "Mean v-component of 10m wind for region Z8 (ms^(-1))",  
                "Mean 10m wind speed for region Z8 (ms^(-1))",  
                "Mean 2m temperature for region Z8 (deg C)",  
                "Mean 2m dewpoint temperature for region Z8 (deg C)")
#
# The next set of predictors is based on weather types. For this purpose 
# the starting point is the "GrossWetterTypes" classification for the 
# Iberia / Western Mediterranean region defined by COST Action 733. However,
# there are eight states in this classification which will probably lead
# to rather complex models. Do some hierarchical clustering therefore,
# to aggregate these states to a more manageable number. The idea is:
# look at proportion of wet days, mean rainfall intensity when wet, and 
# mean temperature, as criteria for the clustering; but in all cases, 
# remve the seasonal cycle first (predominantly motivated by the fact
# that we want the aggregation of states to provide information that is
# relevant after seasonality has been taken into account)
#
GWT.data <- read.table("COST733//GWT08_YR_S01_SP_D09.txt",header=FALSE)
names(GWT.data) <- c("Year","Month","Day","State")
GWT.data <- merge(GWT.data,all.data[,-4])
GWT.pwet <- tapply(resid(lm(as.numeric(RR>0) ~ as.factor(Month),data=GWT.data)),
                   INDEX=list(GWT.data$State[!is.na(GWT.data$RR)]),FUN=mean)
GWT.cmean <- tapply(resid(lm(RR ~ as.factor(Month),subset=(RR>0),data=GWT.data)),
                   INDEX=list(GWT.data$State[GWT.data$RR>0 & !is.na(GWT.data$RR)]),
                   FUN=mean)
GWT.tmean <- tapply(resid(lm(TG ~ as.factor(Month),data=GWT.data)),
                    INDEX=list(GWT.data$State[!is.na(GWT.data$TG)]),
                    FUN=mean)
#
# Standardise so that temperature, rainfall occurrence and rainfall intensity
# are "equally important" in defining clusters
#
GWT.scores <- cbind(scale(GWT.pwet),scale(GWT.cmean),scale(GWT.tmean))
GWT.tree <- hclust(dist(GWT.scores))
x11(width=6,height=6)
plot(GWT.tree,xlab="GWT class",sub="")
dev.copy(pdf,"dendrogram.pdf",width=6,height=6)
dev.off(); dev.off()
#
# Combining the dendrogram with physical intuition (i.e. that groups should pool
# directions that are similar), that suggests pooling groups (1,2), (3,4) and
# (6,7), leaving 5 and 8 on their own. This produces five groups in total, so the
# flexibility is at least comparable to that of the "continuous predictor" approach. 
#
new.groups <- cutree(GWT.tree,k=5)
GWT.data$NewState <- (1:8)[new.groups[GWT.data$State]]
print(table(GWT.data$State,GWT.data$NewState))
#
# So: the new state 1 is "N-NE", state 2 is "E-SE", state 3 is "S", state 4 is
# "W-SW" and state 5 is "NW". 
#
GWT.data <- GWT.data[,-(5:6)]
GWT.data <-aggregate(GWT.data,by=list(GWT.data$Day,GWT.data$Month,GWT.data$Year),
                 FUN=function(x) {x[1]})[,-c(1:3,7)]
GWT.data <- cbind(GWT.data,model.matrix(lm(Year~as.factor(NewState)-1,data=GWT.data)))
GWT.data <- GWT.data[,-4]
daily.preds <- merge(daily.preds,GWT.data)
daily.preds <- daily.preds[order(daily.preds$Year,daily.preds$Month,daily.preds$Day),]

write.GLCexternal(daily.preds,
                  var.names=c(long.names,"N/NE circulation","E/SE circulation",
                              "S circulation","W/SW circulation","NW circulation"),
                  file="DailyPredictors.dat")
#
# Finally: aggregate the weather state information to a monthly scale, merge
# with the previous monthly predictors and output
#
GWT.monthly <- aggregate(GWT.data,by=list(GWT.data$Month,GWT.data$Year),FUN=mean)[,-c(1:2,5)]
monthly.preds <- merge(monthly.preds,GWT.monthly)
monthly.preds <- monthly.preds[order(monthly.preds$Year,monthly.preds$Month),]
write.GLCexternal(as.data.frame(monthly.preds),
                  var.names=c(long.names,"N/NE circulation tendency",
                              "E/SE circulation tendency",
                              "S circulation tendency",
                              "W/SW circulation tendency",
                              "NW circulation tendency"),
                  file="MonthlyPredictors.dat")
